@@ -147,172 +147,16 @@ hide_password (char *url, int len)
 	astfill(second_colon + 1, at_sign - 1);
 }
 
-/* {{{ arginfo */
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_void, 0)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_state_init, 0, 0, 1)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, workgroup)
-	ZEND_ARG_INFO(0, user)
-	ZEND_ARG_INFO(0, password)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_state, 0)
-	ZEND_ARG_INFO(0, state)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_option_get, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, option)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_option_set, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, option)
-	ZEND_ARG_INFO(0, value)
-ZEND_END_ARG_INFO()
-
-#if HAVE_SMBC_SETOPTIONPROTOCOLS
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_client_protocols, 0, 0, 1)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, minproto)
-	ZEND_ARG_INFO(0, maxproto)
-ZEND_END_ARG_INFO()
-#endif
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_path, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_rename, 0)
-	ZEND_ARG_INFO(0, oldstate)
-	ZEND_ARG_INFO(0, oldpath)
-	ZEND_ARG_INFO(0, oldstate)
-	ZEND_ARG_INFO(0, oldpath)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_dir, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, dir)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_file, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, file)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_read, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, file)
-	ZEND_ARG_INFO(0, count)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_open, 0, 0, 3)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-	ZEND_ARG_INFO(0, flags)
-	ZEND_ARG_INFO(0, mode)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_creat, 0, 0, 2)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-	ZEND_ARG_INFO(0, mode)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_write, 0, 0, 3)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, file)
-	ZEND_ARG_INFO(0, buffer)
-	ZEND_ARG_INFO(0, count)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_lseek, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, file)
-	ZEND_ARG_INFO(0, offset)
-	ZEND_ARG_INFO(0, whence)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_ftruncate, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, file)
-	ZEND_ARG_INFO(0, offset)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_utimes, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-	ZEND_ARG_INFO(0, mtime)
-	ZEND_ARG_INFO(0, atime)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO(arginfo_smbclient_getxattr, 0)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-	ZEND_ARG_INFO(0, name)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_smbclient_setxattr, 0, 0, 4)
-	ZEND_ARG_INFO(0, state)
-	ZEND_ARG_INFO(0, path)
-	ZEND_ARG_INFO(0, name)
-	ZEND_ARG_INFO(0, value)
-	ZEND_ARG_INFO(0, flags)
-ZEND_END_ARG_INFO()
-
-/* }}} */
-
-static zend_function_entry smbclient_functions[] =
-{
-	PHP_FE(smbclient_version, arginfo_smbclient_void)
-	PHP_FE(smbclient_library_version, arginfo_smbclient_void)
-	PHP_FE(smbclient_state_new, arginfo_smbclient_void)
-	PHP_FE(smbclient_state_init, arginfo_smbclient_state_init)
-	PHP_FE(smbclient_state_errno, arginfo_smbclient_state)
-	PHP_FE(smbclient_state_free, arginfo_smbclient_state)
-	PHP_FE(smbclient_option_get, arginfo_smbclient_option_get)
-	PHP_FE(smbclient_option_set, arginfo_smbclient_option_set)
-#if HAVE_SMBC_SETOPTIONPROTOCOLS
-	PHP_FE(smbclient_client_protocols, arginfo_smbclient_client_protocols)
-#endif
-	PHP_FE(smbclient_opendir, arginfo_smbclient_path)
-	PHP_FE(smbclient_readdir, arginfo_smbclient_dir)
-	PHP_FE(smbclient_closedir, arginfo_smbclient_dir)
-	PHP_FE(smbclient_stat, arginfo_smbclient_path)
-	PHP_FE(smbclient_fstat, arginfo_smbclient_file)
-	PHP_FE(smbclient_open, arginfo_smbclient_open)
-	PHP_FE(smbclient_creat, arginfo_smbclient_creat)
-	PHP_FE(smbclient_read, arginfo_smbclient_read)
-	PHP_FE(smbclient_close, arginfo_smbclient_file)
-	PHP_FE(smbclient_mkdir, arginfo_smbclient_creat)
-	PHP_FE(smbclient_rmdir, arginfo_smbclient_path)
-	PHP_FE(smbclient_rename, arginfo_smbclient_rename)
-	PHP_FE(smbclient_write, arginfo_smbclient_write)
-	PHP_FE(smbclient_unlink, arginfo_smbclient_path)
-	PHP_FE(smbclient_lseek, arginfo_smbclient_lseek)
-	PHP_FE(smbclient_ftruncate, arginfo_smbclient_ftruncate)
-	PHP_FE(smbclient_chmod, arginfo_smbclient_creat)
-	PHP_FE(smbclient_utimes, arginfo_smbclient_utimes)
-	PHP_FE(smbclient_listxattr, arginfo_smbclient_path)
-	PHP_FE(smbclient_getxattr, arginfo_smbclient_getxattr)
-	PHP_FE(smbclient_setxattr, arginfo_smbclient_setxattr)
-	PHP_FE(smbclient_removexattr, arginfo_smbclient_getxattr)
-	PHP_FE(smbclient_statvfs, arginfo_smbclient_path)
-	PHP_FE(smbclient_fstatvfs, arginfo_smbclient_file)
-#ifdef PHP_FE_END
-	PHP_FE_END
+#if PHP_VERSION_ID < 80000
+#include "smbclient_legacy_arginfo.h"
 #else
-	{NULL, NULL, NULL}
+#include "smbclient_arginfo.h"
 #endif
-};
 
 zend_module_entry smbclient_module_entry =
 	{ STANDARD_MODULE_HEADER
 	, "smbclient"		/* name                  */
-	, smbclient_functions	/* functions             */
+	, ext_functions		/* functions             */
 	, PHP_MINIT(smbclient)		/* module_startup_func   */
 	, PHP_MSHUTDOWN(smbclient)	/* module_shutdown_func  */
 	, PHP_RINIT(smbclient)		/* request_startup_func  */
@@ -1658,15 +1502,21 @@ PHP_FUNCTION(smbclient_listxattr)
 	RETURN_FALSE;
 }
 
+
+/* loop from 16K to 256M */
+#define DEFAULT_BUFFER_SIZE   (16 << 10)
+#define MAXIMUM_BUFFER_SIZE  (256 << 20)
+
 PHP_FUNCTION(smbclient_getxattr)
 {
 	char *url, *name;
 	strsize_t url_len, name_len;
 	int retsize;
-	char values[1000];
+	int xattr_size;
 	zval *zstate;
 	smbc_getxattr_fn smbc_getxattr;
 	php_smbclient_state *state;
+	char *values = NULL;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zstate, &url, &url_len, &name, &name_len) == FAILURE) {
 		return;
@@ -1676,27 +1526,62 @@ PHP_FUNCTION(smbclient_getxattr)
 	if ((smbc_getxattr = smbc_getFunctionGetxattr(state->ctx)) == NULL) {
 		RETURN_FALSE;
 	}
-	/* TODO: 1000 chars should be enough for everyone...?
-	 * However, doing an initial blank call to determine the response size
-	 * seems wasteful, and vulnerable to a time-of-check, time-of-use
-	 * error. */
-	if ((retsize = smbc_getxattr(state->ctx, url, name, values, sizeof(values))) >= 0) {
-		if (retsize > sizeof(values)) {
-			retsize = sizeof(values);
-		}
-#if PHP_MAJOR_VERSION >= 7
-		RETURN_STRINGL(values, retsize);
-#else
-		RETURN_STRINGL(values, retsize, 1);
-#endif
+
+	xattr_size = smbc_getxattr(state->ctx, url, name, NULL, 0);
+
+	if (xattr_size < 0) {
+		goto fail;
 	}
+
+	if (xattr_size == 0) {
+		/* since version 4.16.9 and 4.17.5 this means success :(
+		 * so there is no way to compute the buffer size
+		 * see https://bugzilla.samba.org/show_bug.cgi?id=14808
+		 */
+		xattr_size = DEFAULT_BUFFER_SIZE;
+		do {
+			if (values) {
+				efree(values);
+				xattr_size *= 4;
+			}
+			values = emalloc(xattr_size + 1);
+			retsize = smbc_getxattr(state->ctx, url, name, values, xattr_size + 1);
+		} while (retsize < 0 && xattr_size < MAXIMUM_BUFFER_SIZE);
+	} else {
+		values = emalloc(xattr_size + 1);
+		retsize = smbc_getxattr(state->ctx, url, name, values, xattr_size + 1);
+	}
+
+	if (retsize == 0) { /* success, since 4.16.9 and 4.17.5 */
+		retsize = strlen(values);
+	} else if (retsize > xattr_size) { /* time-of-check, time-of-use error, never happen as recent versions return -1 */
+		retsize = xattr_size;
+	} else if (retsize < 0) {
+		efree(values);
+		goto fail;
+	}
+	/* realloc the string to its real size */
+#if PHP_MAJOR_VERSION >= 7
+	RETVAL_STRINGL(values, retsize);
+#else
+	RETVAL_STRINGL(values, retsize, 1);
+#endif
+	efree(values);
+	return;
+
+fail:
 	hide_password(url, url_len);
 	switch (state->err = errno) {
 		case EINVAL: php_error(E_WARNING, "Couldn't get xattr for %s: library not initialized or incorrect parameter", url); break;
 		case ENOMEM: php_error(E_WARNING, "Couldn't get xattr for %s: out of memory", url); break;
 		case EPERM: php_error(E_WARNING, "Couldn't get xattr for %s: permission denied", url); break;
 		case ENOTSUP: php_error(E_WARNING, "Couldn't get xattr for %s: file system does not support extended attributes", url); break;
-		default: php_error(E_WARNING, "Couldn't get xattr for %s: unknown error (%d)", url, errno); break;
+		default:
+			if (xattr_size == MAXIMUM_BUFFER_SIZE) {
+				php_error(E_WARNING, "Couldn't get xattr for %s: internal buffer is too small", url); break;
+			} else {
+				php_error(E_WARNING, "Couldn't get xattr for %s: unknown error (%d)", url, errno); break;
+			}
 	}
 	RETURN_FALSE;
 }
